from . import base

class Strip(base.Function):
	"""
	Creates a string by concatenating all arguments,
	stripping any leading or trailing whitespace from
	the result.

	====
	The following returns `foo bar`:
	----
	${f:strip:  foo :bar  }
	----
	====
	"""
	def __init__(self):
		# unlimited number of arguments, min 1 argument
		super(Strip, self).__init__(0, 1)

	def execute(self, args):
		if not super(Strip, self).execute(args):
			return None
		return "".join(args).strip()



from . import base

class Exec(base.Function):
	"""
	Executes a process and returns its output.

	====
	The following executes `cat /etc/tuned/active_profile`:
	----
	${f:exec:cat:/etc/tuned/active_profile}
	----
	====
	"""
	def __init__(self):
		# unlimited number of arguments, min 1 argument (the name of executable)
		super(Exec, self).__init__(0, 1)

	def execute(self, args):
		if not super(Exec, self).execute(args):
			return None
		(ret, out) = self._cmd.execute(args)
		if ret == 0:
			return out
		return None



from . import base

class CPUList2HexInvert(base.Function):
	"""
	Converts a CPU list into a hexadecimal CPU mask and inverts it.

	====
	On a system with 4 CPUs numbered from 0 to 3, the following will
	return `00000002`, because only CPU1 is in the complement.
	----
	${f:cpulist2hex_invert:0,2,3}
	----
	====
	"""
	def __init__(self):
		# arbitrary number of arguments
		super(CPUList2HexInvert, self).__init__(0)

	def execute(self, args):
		if not super(CPUList2HexInvert, self).execute(args):
			return None
		# current implementation inverts the CPU list and then converts it to hexmask
		return self._cmd.cpulist2hex(",".join(str(v) for v in self._cmd.cpulist_invert(",,".join(args))))



from . import base

class CPUListInvert(base.Function):
	"""
	Inverts a CPU list, i.e., returns its complement. The complement is
	computed from the list of online CPUs in `/sys/devices/system/cpu/online`.

	====
	On a system with 4 CPUs numbered from 0 to 3, the following will return `1`.
	----
	${f:cpulist_invert:0,2,3}
	----
	====
	"""
	def __init__(self):
		# arbitrary number of arguments
		super(CPUListInvert, self).__init__(0)

	def execute(self, args):
		if not super(CPUListInvert, self).execute(args):
			return None
		return ",".join(str(v) for v in self._cmd.cpulist_invert(",,".join(args)))



from . import base

PROCESSOR_NAME = ("sandybridge",
        "ivybridge",
        "haswell",
        "broadwell",
        "skylake")

PMU_PATH = "/sys/devices/cpu/caps/pmu_name"
ACTIVE = "active"
DISABLE = "disable"

class IntelRecommendedPState(base.Function):
    """
    Returns the recommended intel_pstate CPUFreq driver mode
    based on the CPU generation.

    NOTE: Intel recommends to use the P-State driver
    in active mode with HWP enabled starting from the Ice Lake
    CPU generations. In older CPU generations, setting
    P-State to `active` can introduce jitters which were historically
    seen around and tested with RHEL-7.4. Beginning with the Ice Lake
    generation, Intel has fixed these issues.
    """
    def __init__(self):
        super(IntelRecommendedPState, self).__init__(0)

    def execute(self, args):
        if not super(IntelRecommendedPState, self).execute(args):
            return None

        current_processor_name = self._cmd.read_file(PMU_PATH).strip()
        if current_processor_name == "" or current_processor_name in PROCESSOR_NAME:
            return DISABLE
        return ACTIVE



import re
from . import base

class RegexSearchTernary(base.Function):
	"""
	Ternary regex operator.

	It takes arguments in the following form:
	`STR1, REGEX, STR2, STR3`

	If `REGEX` is matched within `STR1`, returns `STR2`, otherwise returns `STR3`.
	"""
	def __init__(self):
		# 4 arguments
		super(RegexSearchTernary, self).__init__(4, 4)

	def execute(self, args):
		if not super(RegexSearchTernary, self).execute(args):
			return None
		if re.search(args[1], args[0]):
			return args[2]
		else:
			return args[3]



from . import base

class CPUListPack(base.Function):
	"""
	Packs a CPU list into the most succint form.

	====
	The following returns `1-3,5`:
	----
	${f:cpulist_pack:1,2,3,5}
	----
	====
	"""
	def __init__(self):
		# arbitrary number of arguments
		super(CPUListPack, self).__init__(0)

	def execute(self, args):
		if not super(CPUListPack, self).execute(args):
			return None
		return ",".join(str(v) for v in self._cmd.cpulist_pack(",,".join(args)))



from . import base

class S2KB(base.Function):
	"""
	Converts disk sectors to kilobytes.
	"""
	def __init__(self):
		# 1 argument
		super(S2KB, self).__init__(1, 1)

	def execute(self, args):
		if not super(S2KB, self).execute(args):
			return None
		try:
			return str(int(round(int(args[0]) / 2)))
		except ValueError:
			return None



from . import base

class Hex2CPUList(base.Function):
	"""
	Converts a hexadecimal CPU mask into a CPU list.

	====
	The following will return `0,1,2,3`:
	----
	${f:hex2cpulist:00000007}
	----
	====
	"""
	def __init__(self):
		# 1 argument
		super(Hex2CPUList, self).__init__(1, 1)

	def execute(self, args):
		if not super(Hex2CPUList, self).execute(args):
			return None
		return ",".join(str(v) for v in self._cmd.hex2cpulist(args[0]))



import tuned.logs
from . import base
from tuned.profiles.exceptions import InvalidProfileException

log = tuned.logs.get()

class Assertion(base.Function):
	"""
	Compares the second argument and the third argument.
	If they _do not match_, the function logs the text from
	the first argument as an error and aborts profile loading.

	====
	The following will log the error message `fatal error`
	and abort profile loading:
	----
	${f:assertion:fatal error:3:5}
	----
	====
	"""
	def __init__(self):
		# 3 arguments
		super(Assertion, self).__init__(3, 3)

	def execute(self, args):
		if not super(Assertion, self).execute(args):
			return None
		if args[1] != args[2]:
			log.error("assertion '%s' failed: '%s' != '%s'" % (args[0], args[1], args[2]))
			raise InvalidProfileException("Assertion '%s' failed." % args[0])
		return None



from . import base

class VirtCheck(base.Function):
	"""
	Checks whether *TuneD* is running inside a virtual machine (VM) or on bare metal.

	Inside a VM, it returns the first argument.
	Otherwise returns the second argument (even on error).

	====
	The following returns `VM` when running in a virtual machine:
	----
	${f:virt_check:VM:Bare}
	----
	====
	"""
	def __init__(self):
		# 2 arguments
		super(VirtCheck, self).__init__(2, 2)

	def execute(self, args):
		if not super(VirtCheck, self).execute(args):
			return None
		(ret, out) = self._cmd.execute(["virt-what"])
		if ret == 0 and len(out) > 0:
			return args[0]
		return args[1]



import re
import tuned.logs
from . import base

log = tuned.logs.get()

class LSCPUCheck(base.Function):
	"""
	Checks regexes against the output of `lscpu`.

	Accepts arguments in the form `REGEX1, STR1, REGEX2, STR2, ...[, STR_FALLBACK]`.

	If `REGEX1` has a match in the output of `lscpu`, it returns `STR1`.
	
	If `REGEX2` has a match, it returns `STR2`.

	The function stops on the first match, i.e., if `REGEX1` has a match,
	no more regexes are processed. If no regex has a match, `STR_FALLBACK`
	is returned. If there is no fallback value, returns an empty string.
	"""
	def __init__(self):
		# unlimited number of arguments, min 2 arguments
		super(LSCPUCheck, self).__init__(0, 2)

	def execute(self, args):
		if not super(LSCPUCheck, self).execute(args):
			return None
		# Stdout is the 2nd result from the execute call
		_, lscpu = self._cmd.execute("lscpu")
		for i in range(0, len(args), 2):
			if i + 1 < len(args):
				try:
					if re.search(args[i], lscpu, re.MULTILINE):
						return args[i + 1]
				except re.error:
					log.error("Invalid regular expression: '%s'" % args[i])
		if len(args) % 2:
			return args[-1]
		else:
			return ""



import re
from . import base

class CPUInfoCheck(base.Function):
	"""
	Checks regexes against the content of `/proc/cpuinfo`.

	Accepts arguments in the form `REGEX1, STR1, REGEX2, STR2, ...[, STR_FALLBACK]`.

	If `REGEX1` has a match in `/proc/cpuinfo`, it returns `STR1`.

	If `REGEX2` has a match, it returns `STR2`.

	The function stops on the first match, i.e., if `REGEX1` has a match,
	no more regexes are processed. If no regex has a match, `STR_FALLBACK`
	is returned. If there is no fallback value, it returns an empty string.
	"""
	def __init__(self):
		# unlimited number of arguments, min 2 arguments
		super(CPUInfoCheck, self).__init__(0, 2)

	def execute(self, args):
		if not super(CPUInfoCheck, self).execute(args):
			return None
		cpuinfo = self._cmd.read_file("/proc/cpuinfo")
		for i in range(0, len(args), 2):
			if i + 1 < len(args):
				if re.search(args[i], cpuinfo, re.MULTILINE):
					return args[i + 1]
		if len(args) % 2:
			return args[-1]
		else:
			return ""



from . import base

class CPUListPresent(base.Function):
	"""
	Checks whether the CPUs from a given CPU list are present on the system.
	Returns a CPU list containing only the present CPUs from the given list.
	"""
	def __init__(self):
		# arbitrary number of arguments
		super(CPUListPresent, self).__init__(0)

	def execute(self, args):
		if not super(CPUListPresent, self).execute(args):
			return None
		cpus = self._cmd.cpulist_unpack(",,".join(args))
		present = self._cmd.cpulist_unpack(self._cmd.read_file("/sys/devices/system/cpu/present"))
		return ",".join(str(v) for v in sorted(list(set(cpus).intersection(set(present)))))



from . import base

class CPUListUnpack(base.Function):
	"""
	Unpacks a CPU list into a form with no ranges.

	====
	The following returns `1,2,3,5`:
	----
	${f:cpulist_unpack:1-3,5}
	----
	====
	"""
	def __init__(self):
		# arbitrary number of arguments
		super(CPUListUnpack, self).__init__(0)

	def execute(self, args):
		if not super(CPUListUnpack, self).execute(args):
			return None
		return ",".join(str(v) for v in self._cmd.cpulist_unpack(",,".join(args)))



from . import base

class CPUListOnline(base.Function):
	"""
	Returns a CPU list containing the online CPUs from the given CPU list.

	====
	On a system with 8 CPUs where the first 4 CPUs (0 to 3) are offline,
	the following returns `4-6`:
	----
	${f:cpulist_online:0-6}
	----
	====
	"""
	def __init__(self):
		# arbitrary number of arguments
		super(CPUListOnline, self).__init__(0)

	def execute(self, args):
		if not super(CPUListOnline, self).execute(args):
			return None
		cpus = self._cmd.cpulist_unpack(",".join(args))
		online = self._cmd.cpulist_unpack(self._cmd.read_file("/sys/devices/system/cpu/online"))
		return ",".join(str(v) for v in cpus if v in online)



import collections
import os
import errno
import threading
import tuned.logs
from tuned.exceptions import TunedException
from tuned.profiles.exceptions import InvalidProfileException
import tuned.consts as consts
from tuned.utils.commands import commands
from tuned import exports
from tuned.utils.profile_recommender import ProfileRecommender
import re

log = tuned.logs.get()


class Daemon(object):
	def __init__(self, unit_manager, profile_loader, profile_names=None, config=None, application=None):
		log.debug("initializing daemon")
		self._daemon = consts.CFG_DEF_DAEMON
		self._sleep_interval = int(consts.CFG_DEF_SLEEP_INTERVAL)
		self._update_interval = int(consts.CFG_DEF_UPDATE_INTERVAL)
		self._dynamic_tuning = consts.CFG_DEF_DYNAMIC_TUNING
		self._recommend_command = True
		self._rollback = consts.CFG_DEF_ROLLBACK
		if config is not None:
			self._daemon = config.get_bool(consts.CFG_DAEMON, consts.CFG_DEF_DAEMON)
			self._sleep_interval = int(config.get(consts.CFG_SLEEP_INTERVAL, consts.CFG_DEF_SLEEP_INTERVAL))
			self._update_interval = int(config.get(consts.CFG_UPDATE_INTERVAL, consts.CFG_DEF_UPDATE_INTERVAL))
			self._dynamic_tuning = config.get_bool(consts.CFG_DYNAMIC_TUNING, consts.CFG_DEF_DYNAMIC_TUNING)
			self._recommend_command = config.get_bool(consts.CFG_RECOMMEND_COMMAND, consts.CFG_DEF_RECOMMEND_COMMAND)
			self._rollback = config.get(consts.CFG_ROLLBACK, consts.CFG_DEF_ROLLBACK)
		self._application = application
		if self._sleep_interval <= 0:
			self._sleep_interval = int(consts.CFG_DEF_SLEEP_INTERVAL)
		if self._update_interval == 0:
			self._dynamic_tuning = False
		elif self._update_interval < self._sleep_interval:
			self._update_interval = self._sleep_interval
		self._sleep_cycles = self._update_interval // self._sleep_interval
		log.info("using sleep interval of %d second(s)" % self._sleep_interval)
		if self._dynamic_tuning:
			log.info("dynamic tuning is enabled (can be overridden by plugins)")
			log.info("using update interval of %d second(s) (%d times of the sleep interval)" % (self._sleep_cycles * self._sleep_interval, self._sleep_cycles))

		self._profile_recommender = ProfileRecommender(is_hardcoded = not self._recommend_command)
		self._unit_manager = unit_manager
		self._profile_loader = profile_loader
		self._init_threads()
		self._cmd = commands()
		try:
			self._init_profile(profile_names)
		except TunedException as e:
			log.error("Cannot set initial profile. No tunings will be enabled: %s" % e)
			if not self._daemon:
				raise TunedException("Applying TuneD profile failed, check TuneD logs for details.")

	def _init_threads(self):
		self._thread = None
		self._terminate = threading.Event()
		# Flag which is set if terminating due to profile_switch
		self._terminate_profile_switch = threading.Event()
		# Flag which is set if there is no operation in progress
		self._not_used = threading.Event()
		# Flag which is set if SIGHUP is being processed
		self._sighup_processing = threading.Event()
		# Flag which is set if there is unprocessed SIGHUP pending
		self._sighup_pending = threading.Event()
		self._not_used.set()
		self._profile_applied = threading.Event()

	def reload_profile_config(self):
		"""Read configuration files again and load profile according to them"""
		self._init_profile(None)

	def _init_profile(self, profile_names):
		manual = True
		post_loaded_profile = self._cmd.get_post_loaded_profile()
		if profile_names is None:
			(profile_names, manual) = self._get_startup_profile()
			if profile_names is None:
				msg = "No profile is preset, running in manual mode. "
				if post_loaded_profile:
					msg += "Only post-loaded profile will be enabled"
				else:
					msg += "No profile will be enabled."
				log.info(msg)
		# Passed through '-p' cmdline option
		elif profile_names == "":
			if post_loaded_profile:
				log.info("Only post-loaded profile will be enabled")
			else:
				log.info("No profile will be enabled.")

		self._profile = None
		self._manual = None
		self._active_profiles = []
		self._post_loaded_profile = None
		self.set_all_profiles(profile_names, manual, post_loaded_profile)

	def _load_profiles(self, profile_names, manual):
		profile_names = profile_names or ""
		profile_list = profile_names.split()

		if self._post_loaded_profile:
			log.info("Using post-loaded profile '%s'"
				 % self._post_loaded_profile)
			profile_list = profile_list + [self._post_loaded_profile]
		for profile in profile_list:
			if profile not in self.profile_loader.profile_locator.get_known_names():
				errstr = "Requested profile '%s' doesn't exist." % profile
				self._notify_profile_changed(profile_names, False, errstr)
				raise TunedException(errstr)
		try:
			if profile_list:
				self._profile = self._profile_loader.load(profile_list)
			else:
				self._profile = None

			self._manual = manual
			self._active_profiles = profile_names.split()
		except InvalidProfileException as e:
			errstr = "Cannot load profile(s) '%s': %s" % (" ".join(profile_list), e)
			self._notify_profile_changed(profile_names, False, errstr)
			raise TunedException(errstr)

	def set_profile(self, profile_names, manual):
		if self.is_running():
			errstr = "Cannot set profile while the daemon is running."
			self._notify_profile_changed(profile_names, False,
						     errstr)
			raise TunedException(errstr)

		self._load_profiles(profile_names, manual)

	def _set_post_loaded_profile(self, profile_name):
		if not profile_name:
			self._post_loaded_profile = None
		elif len(profile_name.split()) > 1:
			errstr = "Whitespace is not allowed in profile names; only a single post-loaded profile is allowed."
			raise TunedException(errstr)
		else:
			self._post_loaded_profile = profile_name

	def set_all_profiles(self, active_profiles, manual, post_loaded_profile,
			     save_instantly=False):
		if self.is_running():
			errstr = "Cannot set profile while the daemon is running."
			self._notify_profile_changed(active_profiles, False,
						     errstr)
			raise TunedException(errstr)

		self._set_post_loaded_profile(post_loaded_profile)
		self._load_profiles(active_profiles, manual)

		if save_instantly:
			self._save_active_profile(active_profiles, manual)
			self._save_post_loaded_profile(post_loaded_profile)

	@property
	def profile(self):
		return self._profile

	@property
	def manual(self):
		return self._manual

	@property
	def post_loaded_profile(self):
		# Return the profile name only if the profile is active. If
		# the profile is not active, then the value is meaningless.
		return self._post_loaded_profile if self._profile else None

	@property
	def profile_recommender(self):
		return self._profile_recommender

	@property
	def profile_loader(self):
		return self._profile_loader

	# send notification when profile is changed (everything is setup) or if error occured
	# result: True - OK, False - error occured
	def _notify_profile_changed(self, profile_names, result, errstr):
		if self._application is not None:
			exports.send_signal(consts.SIGNAL_PROFILE_CHANGED, profile_names, result, errstr)
		return errstr

	def _full_rollback_required(self):
		retcode, out = self._cmd.execute(["systemctl", "is-system-running"], no_errors = [0])
		if retcode < 0:
			return False
		if out[:8] == "stopping":
			return False
		retcode, out = self._cmd.execute(["systemctl", "list-jobs"], no_errors = [0])
		return re.search(r"\b(shutdown|reboot|halt|poweroff)\.target.*start", out) is None and not retcode

	def _ensure_bootloader_unit(self):
		"""Ensure the bootloader plugin is loaded on bootc systems.

		On systems where bootc loader-entries set-options-for-source is
		available, the bootloader plugin must always be loaded - even
		for profiles without a [bootloader] section.  Without it, stale
		kargs from a previous profile cannot be cleared during a
		profile switch because the plugin (and its unapply logic) would
		never be instantiated.
		"""
		if "bootloader" in self._profile.units:
			return
		(retcode, out, err) = self._cmd.execute(
			["bootc", "loader-entries", "set-options-for-source",
			 "--help"], return_err=True)
		if retcode != 0:
			return
		from tuned.profiles.unit import Unit
		log.debug("injecting synthetic bootloader unit for bootc source tracking")
		self._profile.units["bootloader"] = Unit(
			"bootloader", collections.OrderedDict())

	def _thread_code(self):
		if self._profile is None:
			raise TunedException("Cannot start the daemon without setting a profile.")

		self._ensure_bootloader_unit()
		self._unit_manager.create(self._profile.units)
		self._save_active_profile(" ".join(self._active_profiles),
					  self._manual)
		self._save_post_loaded_profile(self._post_loaded_profile)
		self._unit_manager.start_tuning()
		self._profile_applied.set()
		log.info("static tuning from profile '%s' applied" % self._profile.name)
		if self._daemon:
			exports.start()
		profile_names = " ".join(self._active_profiles)
		self._notify_profile_changed(profile_names, True, "OK")
		self._sighup_processing.clear()

		if self._daemon:
			# In python 2 interpreter with applied patch for rhbz#917709 we need to periodically
			# poll, otherwise the python will not have chance to update events / locks (due to GIL)
			# and e.g. DBus control will not work. The polling interval of 1 seconds (which is
			# the default) is still much better than 50 ms polling with unpatched interpreter.
			# For more details see TuneD rhbz#917587.
			_sleep_cnt = self._sleep_cycles
			while not self._cmd.wait(self._terminate, self._sleep_interval):
				if self._dynamic_tuning:
					_sleep_cnt -= 1
					if _sleep_cnt <= 0:
						_sleep_cnt = self._sleep_cycles
						log.debug("updating monitors")
						self._unit_manager.update_monitors()
						log.debug("performing tunings")
						self._unit_manager.update_tuning()

		self._profile_applied.clear()

		# wait for others to complete their tasks, use timeout 3 x sleep_interval to prevent
		# deadlocks
		i = 0
		while not self._cmd.wait(self._not_used, self._sleep_interval) and i < 3:
			i += 1

		# if terminating due to profile switch
		if self._terminate_profile_switch.is_set():
			rollback = consts.ROLLBACK_FULL
		else:
			# Assume only soft rollback is needed. Soft rollback means reverting all
			# non-persistent tunings applied by a plugin instance. In contrast to full
			# rollback, information about what to revert is kept in RAM (volatile
			# memory) -- TuneD data structures.
			# With systemd TuneD detects system shutdown and in such a case it doesn't
			# perform full cleanup. If the system is not shutting down, it means that TuneD
			# was explicitly stopped by the user and in such case do the full cleanup. On
			# systems without systemd, full cleanup is never performed.
			rollback = consts.ROLLBACK_SOFT
			if not self._full_rollback_required():
				log.info("terminating TuneD due to system shutdown / reboot")
			elif self._rollback == "not_on_exit":
				# no rollback on TuneD exit whatsoever
				rollback = consts.ROLLBACK_NONE
				log.info("terminating TuneD and not rolling back any changes due to '%s' option in '%s'" % (consts.CFG_ROLLBACK, consts.GLOBAL_CONFIG_FILE))
			else:
				if self._daemon:
					log.info("terminating TuneD, rolling back all changes")
					rollback = consts.ROLLBACK_FULL
				else:
					log.info("terminating TuneD in one-shot mode")
		if self._daemon:
			self._unit_manager.stop_tuning(rollback)
		self._unit_manager.destroy_all()

	def _save_active_profile(self, profile_names, manual):
		try:
			self._cmd.save_active_profile(profile_names, manual)
		except TunedException as e:
			log.error(str(e))

	def _save_post_loaded_profile(self, profile_name):
		try:
			self._cmd.save_post_loaded_profile(profile_name)
		except TunedException as e:
			log.error(str(e))

	def _get_recommended_profile(self):
		log.info("Running in automatic mode, checking what profile is recommended for your configuration.")
		profile = self._profile_recommender.recommend()
		log.info("Using '%s' profile" % profile)
		return profile

	def _get_startup_profile(self):
		profile, manual = self._cmd.get_active_profile()
		if manual is None:
			manual = profile is not None
		if not manual:
			profile = self._get_recommended_profile()
		return profile, manual

	def get_all_plugins(self):
		"""Return all accessible plugin classes"""
		return self._unit_manager.plugins_repository.load_all_classes()

	def get_plugin_documentation(self, plugin_name):
		"""Return plugin class docstring"""
		try:
			plugin_class = self._unit_manager.plugins_repository.load_class(
				plugin_name
			)
		except ImportError:
			return ""
		return plugin_class.__doc__

	def get_plugin_hints(self, plugin_name):
		"""Return plugin's parameters and their hints

		Parameters:
		plugin_name -- plugins name

		Return:
		dictionary -- {parameter_name: hint}
		"""
		try:
			plugin_class = self._unit_manager.plugins_repository.load_class(
				plugin_name
			)
		except ImportError:
			return {}
		return plugin_class.get_config_options_hints()

	def is_enabled(self):
		return self._profile is not None

	def is_running(self):
		return self._thread is not None and self._thread.is_alive()

	def start(self):
		if self.is_running():
			return False

		if self._profile is None:
			return False

		log.info("starting tuning")
		self._not_used.set()
		self._thread = threading.Thread(target=self._thread_code)
		self._terminate_profile_switch.clear()
		self._terminate.clear()
		self._thread.start()
		return True

	def verify_profile(self, ignore_missing):
		if not self.is_running():
			log.error("TuneD is not running")
			return False

		if self._profile is None:
			log.error("no profile is set")
			return False

		if not self._profile_applied.is_set():
			log.error("profile is not applied")
			return False

		# using daemon, the main loop mustn't exit before our completion
		self._not_used.clear()
		log.info("verifying profile(s): %s" % self._profile.name)
		ret = self._unit_manager.verify_tuning(ignore_missing)
		# main loop is allowed to exit
		self._not_used.set()
		return ret

	# profile_switch is helper telling plugins whether the stop is due to profile switch
	def stop(self, profile_switch = False):
		if not self.is_running():
			return False
		log.info("stopping tuning")
		if profile_switch:
			self._terminate_profile_switch.set()
		self._terminate.set()
		self._thread.join()
		self._thread = None

		return True



from . import base

class KB2S(base.Function):
	"""
	Converts kilobytes to disk sectors.
	"""
	def __init__(self):
		# 1 argument
		super(KB2S, self).__init__(1, 1)

	def execute(self, args):
		if not super(KB2S, self).execute(args):
			return None
		try:
			return str(int(args[0]) * 2)
		except ValueError:
			return None



import tuned.logs
from . import base

log = tuned.logs.get()

class Log(base.Function):
	"""
	Returns the concatenation of its arguments and also logs the return value,
	which is useful for debugging.

	.Using `log` to debug intermediate values
	====
	Since the arguments of `log` "fall through" the function, it
	can be used as below for debugging intermediate values:
	----
	[variables]
	isolated_cores_hex = ${f:cpulist2hex:${f:log:${f:calc_isolated_cores}}}
	----
	====
	"""
	def __init__(self):
		# unlimited number of arguments, min 1 argument (the value to log)
		super(Log, self).__init__(0, 1)

	def execute(self, args):
		if not super(Log, self).execute(args):
			return None
		s = "".join(args)
		log.info(s)
		return s



from . import base

class CPUList2Hex(base.Function):
	"""
	Converts a CPU list into a hexadecimal CPU mask.

	====
	The following will return `00000007`.
	----
	${f:cpulist2hex:0-3}
	----
	====
	"""
	def __init__(self):
		# arbitrary number of arguments
		super(CPUList2Hex, self).__init__(0)

	def execute(self, args):
		if not super(CPUList2Hex, self).execute(args):
			return None
		return self._cmd.cpulist2hex(",,".join(args))



from . import base

class CPUList2Devs(base.Function):
	"""
	Converts a CPU list into a comma-separated list of device names.

	====
	The following will return `cpu1,cpu2,cpu3,cpu5`:
	----
	${f:cpulist2devs:1-3,5}
	----
	====
	"""
	def __init__(self):
		# arbitrary number of arguments
		super(CPUList2Devs, self).__init__(0)

	def execute(self, args):
		if not super(CPUList2Devs, self).execute(args):
			return None
		return self._cmd.cpulist2string(self._cmd.cpulist_unpack(",".join(args)), prefix = "cpu")
